#include "precomp.h"

#define REGADDRESS2INDEX(addr)  ((addr) - 128)

static CRITICAL_SECTION gs_cs;
static volatile LONG gs_blInitialized;

class CAutoCriticalSection
{
public:
    explicit CAutoCriticalSection(CRITICAL_SECTION* pcs) : m_pcs(pcs) { EnterCriticalSection(m_pcs); }
    ~CAutoCriticalSection() { LeaveCriticalSection(m_pcs); }
private:
    CRITICAL_SECTION* m_pcs;
};

#define MAKE_UVISION_PTR(seg, addr) (UL32)(((UL32)(addr)) | (((UL32)(seg)) << 24U))
#define MAKE_UVISION_CODEPTR(addr)  MAKE_UVISION_PTR(amCODE, (addr))
#define MAKE_UVISION_DATAPTR(addr)  MAKE_UVISION_PTR(amDATA, (addr))
#define MAKE_UVISION_IDATAPTR(addr) MAKE_UVISION_PTR(amIDATA, (addr))
#define MAKE_UVISION_XDATAPTR(addr) MAKE_UVISION_PTR(amXDATA, (addr))

static HRESULT GetDebugAdapterId(ISmartICE04Devices* pDevices, BSTR* pDeviceId);

static RG51         REG51;          // 8051 Registers
static BYTE      PlayDead;          // Disables the driver after comm breaks down.
static BYTE   RegUpToDate;          // Flag: mark driver is 'dead' (disconnected)
static BYTE     FlashProg;          // Debugger started in Flash Programming mode

static OIL    ProgressBar;          // Progress-Bar data
static int        NumRecs;          // number of trace records


/*
 * AGDI-Data
 */

static SUPP             supp;       // supported features
static UC8              iRun;       // target is currently executing
static UC8           StopRun;       // Flag: stop execution
static AG_Bps       **pBhead;
static UL32          *pCURPC;
static UL32            curPC;

static const AG_Bps     *pBX;       // used for temporary Breakpoint
static UL32          abreaks;       // number of code-address Bp's
static UL32          cbreaks;       // number of conditional Bp's
static UL32          wbreaks;       // number of access Bp's
static UL32      SetBrkCount;
// static const char szNBP[] = "*** AGDI-Error: can't setup break at 0x%08X\n";


static UL32           Uv2Msg;       // Uv2-registered Message token
static pCBF          pCbFunc;       // call-back function of s166

static DYMENU Menu[] =  {
//  {  1, "Target Settings", ConfDisp, 0, 0, nullptr },  // modal dialog
//  {  1, "Modeless Dialog",   MdShow, 0, 0, &ModDlg },  // modeless dialog
    { -1, nullptr, nullptr, 0, 0, nullptr }                       // End of menu list
};

// if an extension modeless dialog is activated, then '*pHwnd' *MUST*
// receive the HWND (m_hWnd in MFC) of the dialog.  If the dialog
// looses focus or is closed, then '*pHwnd' *MUST* be set to NULL.
// This is to ensure the proper message chain within Uv2.

static const HWND    *pHwnd;
       HWND           hMfrm;        // parent handle (CMainFrame)
static HMODULE        hInst;        // this DLL's instance handle

/*
 * Register View Management
 */

static void StopTarget();
static void RegGet (RITEM *vp, unsigned nR);
static I32 RegSet (const RITEM *vp, const GVAL *pV);
static ISmartICE04* gs_pProgrammer;
static volatile LONG gs_Running = 0;

static struct
{
    UL32    address;
    BOOL    enabled;
} gs_Breakpoints[] =
{
    { 0, FALSE },
    { 0, FALSE },
    { 0, FALSE },
    { 0, FALSE }
};

/*
 * Initialize Register View
 */

static void InitRegs()  {
  REGDSC    dsc;
  ChipId    id = UnknownChip;

  if(gs_pProgrammer)
    gs_pProgrammer->get_ChipId(&id);

  dsc.GrpArr  = GetRegisterGroups(id, &dsc.nGitems);
  dsc.RegArr  = GetRegisterItems(id, &dsc.nRitems);
  dsc.RegGet  = RegGet;                 // get function
  dsc.RegSet  = RegSet;                 // set function
  pCbFunc (AG_CB_INITREGV, &dsc);       // Install RegView in uVision2
}

/*
 * Initialize uVision2 Progress bar
 */

static void InitProgress (const char *msg)  {
  ProgressBar.Job   = PROGRESS_INIT;
  ProgressBar.pos   = 0;
  ProgressBar.low   = 0;
  ProgressBar.hig   = 100;
  ProgressBar.label = reinterpret_cast<SC8*>(const_cast<char*>(msg));
  pCbFunc (AG_CB_PROGRESS, &ProgressBar);
}

/*
 * Kill uVision2 Progress bar
 */

static void StopProgress()  {
  ProgressBar.Job = PROGRESS_KILL;
  pCbFunc (AG_CB_PROGRESS, &ProgressBar);
}

/*
 * Set Progress bar
 */

static void SetProgress (DWORD pos )  {
  if (pos > 100) pos = 100;
  ProgressBar.Job = PROGRESS_SETPOS;
  ProgressBar.pos = static_cast<I32>(pos);
  pCbFunc (AG_CB_PROGRESS, &ProgressBar);
}


//--- Memory cache handling.
//--- Simulate the (non-existing) target memory
// Memory caching is used to increase debugger performance.
// When uVision2 requests data, it is taken out from cache memory buffer
// instead of reading it from the target hardware. This performance
// is increased slightly if fast interface is used (USB, Ethernet,...)

static struct MMM  mslots [520];    // cache memory slots

/*
 * Free all allocated memory by this driver before driver shutdown.
 */

static void FreeCache()  {
  for(uintptr_t i = 0 ; i < _countof(mslots); ++i )  {
    struct MMM* s = &mslots [i];      // a16-a23
    if (s->mem)  {
      free (s->mem);
      free (s->atr);
      s->mem = nullptr;
      s->atr = nullptr;
    }
  }
//---TODO: free additional memory resources, if allocated
}


/*
 * Map full address to Cache-Slot-number:
 */

static DWORD SlotNo (DWORD n)  {
  switch (n >> 24)  {                 // high byte of address holds mSpace.
    case amDATA:
      if ((n & 0xFF) >= 0x80)  {      // SFR-address
        n = 0x200;                    // sfr slot at 0x200
        break;
      }
    case amIDATA:
      n = 0x201;                      // idata slot at 0x201
      break;
    case amPDATA:
    case amXDATA:
      n = 0x100 + ((n >> 16) & 0xFF); // xdata slots start at 0x0100
      break;
    default:                          // code
      n = (n >> 16) & 0xFF;
      break;
  }
  return (n);
}


/*
 * Cache data
 */

static void MyCacheData(const BYTE *pB, DWORD nAdr, DWORD nCnt)
{
    DWORD  aMask;

    if(!pB)
        return;

    while(nCnt)
    {
        const DWORD nSlot = SlotNo(nAdr);

        switch (nAdr >> 24)
        {
            case amDATA:
            case amPDATA:
            case amIDATA:
                aMask = 0xFF;
                break;
            default:
                aMask = 0xFFFF;
                break;
        }

        U16* pA = mslots [nSlot].atr;
        BYTE* pM = mslots [nSlot].mem;

        if(pA && pM)                            // both of them need to be valid
        {
            pA [nAdr & aMask] |= AG_ATR_UPTD;   // 'up to date' attribute
            pM [nAdr & aMask]  = *pB;           // the data
        }

        ++nAdr;
        ++pB;
        --nCnt;
    }
}


/*
 * clear 'AG_ATR_UPDT' for the given range
 */

static void ClearCaR (DWORD nAdr, DWORD nCnt)
{
    DWORD aMask;

    for ( ; nCnt != 0 ; --nCnt)
    {
        const DWORD nSlot = SlotNo (nAdr);

        switch (nAdr >> 24)
        {
            case amDATA:
            case amPDATA:
            case amIDATA:
                aMask = 0xFF;
                break;
            default:
                aMask = 0xFFFF;
                break;
        }

        U16* const pA = mslots [nSlot].atr;

        if (pA != nullptr) {                     // attributes are valid.
            pA [nAdr & aMask] &= ~AG_ATR_UPTD;   // clear 'up to date' attribute
        }

        ++nAdr;
    }
}



/*
 * Check if cached data is valid.
 */

static int CacheValid (BYTE *pB, DWORD nAdr, DWORD nCnt)  {
  int   n1, n2;
  DWORD aMask;

  if (pB == nullptr) return (0);

  switch (nAdr >> 24) {
    case amDATA:
      if(!(MonConf.Opt & CACHE_DATA))
        return(0);

      if ((nAdr & 0xFF) < 0x80) break;  // Cache internal DATA
      return (0);                       // SFRs are not cached

    case amIDATA:                       // Cache DATA, IDATA
      if (MonConf.Opt & CACHE_DATA) break;
      return (0);

    case amXDATA:                       // Cache XDATA ram
      if (MonConf.Opt & CACHE_XDATA) break;
      return (0);

    case amCODE:                        // Cache CODE mem.
      if (MonConf.Opt & CACHE_CODE) break;
      return (0);

    default:
      return (0);
  }

  n2 = 0;

  for (n1 = static_cast<int>(nCnt); n1; --n1)  {
    const DWORD nSlot = SlotNo (nAdr);
    switch (nAdr >> 24)  {
      case amDATA:
      case amPDATA:
      case amIDATA:
        aMask = 0xFF;
        break;
      default:
        aMask = 0xFFFF;
        break;
    }
    const U16* pA = mslots [nSlot].atr;
    const BYTE* pM = mslots [nSlot].mem;
    if (!pA || !pM) return (0);          // both of them need to be valid
    if (!(pA [nAdr & aMask] & AG_ATR_UPTD))  {
      return (0);                        // cache data invalid.
    }
    pB [n2] = pM [nAdr & aMask];         // copy data to user buffer
    ++n2;
    ++nAdr;
  }
  return (1);                            // cached data valid.
}


/*
 * Map Memory
 */

static void MapSeg (DWORD nSeg, DWORD nSof, DWORD nLen, WORD16 nAtr)  {
  struct MMM   *s;
  WORD16      *pA;
  DWORD         n;

  if (nLen == 0) return;
  n = SlotNo (nSeg);
  s = &mslots [n];
  if (s->mem == nullptr)  {                         // if not already mapped...
    s->mem = static_cast<UC8*>(calloc (MSGM, 1));   // allocate 64k block
    s->atr = static_cast<U16*>(calloc (ASGM, 1));   // and associated attributes
    if ((s->mem == nullptr) || (s->atr == nullptr)){// malloc failed.
      s->mem = nullptr;
      s->atr = nullptr;
      return;
    }
  }

  pA = &s->atr [nSof];
  for ( n = 0 ; n < nLen ; ++n)  {                  // initialize attribs
    *pA &= (AG_ATR_READ | AG_ATR_WRITE | AG_ATR_EXEC | AG_ATR_EXTR | AG_ATR_UPTD);
    *pA++ |= nAtr;
  }
}


/*
 * Map a cache segment descriptor.
 */

static DWORD MMapMem (UL32 nAtr, UL32 nAdr, UL32 nLen)  {
  if (nAtr == 0)  {
    nAtr = AG_ATR_READ | AG_ATR_WRITE | AG_ATR_EXEC;
  }
  MapSeg (nAdr, nAdr & 0xFFFFFF, nLen, static_cast<U16>(nAtr));
  return (0);
}


/*
 * Get Pointer to Attribute Segment for address 'nAdr'
 */

static U16 *MGetAttrSeg (UL32 nAdr)  {
  const DWORD n = SlotNo(nAdr); // map address to cache slot
  return(mslots[n].atr);        // and return the attribute segment.
}


/*
 * Get Pointer to Attributes for address 'nAdr'
 */

static U16 *MGetAttr (UL32 nAdr)  {
  U16* pA = MGetAttrSeg (nAdr); // get attribute segment
  if (pA != nullptr)  {
    pA = &pA [nAdr & 0xFFFF];   // address of attributes for address 'nAdr'
  }
  return (pA);
}

static void InitMem()  {
  MapSeg (amDATA,    0x00,   0x100, AG_ATR_READ | AG_ATR_WRITE);
  MapSeg (amIDATA,   0x80,    0x80, AG_ATR_READ | AG_ATR_WRITE);
  MapSeg (amXDATA, 0x0000, 0x10000, AG_ATR_READ | AG_ATR_WRITE);
  MapSeg (amCODE,  0x0000, 0x10000, AG_ATR_READ | AG_ATR_EXEC);
//---TODO:
//---add cache buffer for additional memory space here if needed
}

//--- End of c/i-chache


/*
 * Output a message line into uVision2's command window in debug mode
 * and to build window in Flash download mode
 */

#pragma warning(push)
#pragma warning(disable: 4996)
static void Debug (const TCHAR *fmt, ...)  {
#ifdef _DEBUG
  va_list   argptr;
  TCHAR     txtbuf [2048];

  va_start(argptr, fmt);
  StringCchVPrintf(&txtbuf[0], _countof(txtbuf), fmt, argptr);
  OutputDebugString(txtbuf);
  va_end(argptr);
#else
  UNREFERENCED_PARAMETER(fmt);
#endif
}
#pragma warning(pop)


static void txtout (const char *fmt, ...)  { // this is a debug 'printf'
  va_list   argptr;
  char      txtbuf [2048];

  va_start(argptr, fmt);
  StringCchVPrintfA(&txtbuf[0], _countof(txtbuf), fmt, argptr);
  SendMessage (hMfrm,                // for Flash Download send message to Build window
               Uv2Msg, FlashProg ? MSG_UV2_TRNLINE : MSG_UV2_CMDLINE,
              reinterpret_cast<LPARAM>(txtbuf));
  va_end(argptr);
}

static void txtout_buildwnd (const char *fmt, ...)  {
  va_list   argptr;
  char      txtbuf [2048];

  va_start(argptr, fmt);
  StringCchVPrintfA(&txtbuf[0], _countof(txtbuf), fmt, argptr);
  SendMessage (hMfrm,                // for Flash Download send message to Build window
               Uv2Msg, MSG_UV2_TRNLINE,
              reinterpret_cast<LPARAM>(txtbuf));
  va_end(argptr);
}



/*
 * This function is required for Remote-Setup in uVision2 Debug mode
 */
//static void ConfDisp (DYMENU *pM)  {
//  struct MonConf oCnf;
//  int     i;
//
//  pM;
//  oCnf = MonConf;              // save current setup
//  i    = DoDlgSetup(nullptr);     // start setup dialog
//
//  if (i == IDOK)  {
//    if (memcmp (&MonConf, &oCnf, sizeof (oCnf)) != 0)  {   // configuration has changed
//      i = MessageBoxA (hMfrm,"Configuration has been changed, take new values ?",
//                             "Target Monitor Notification",
//                             MB_OKCANCEL | MB_ICONWARNING);
//      if (i == IDCANCEL)  {
//        MonConf = oCnf;         // restore previous configuration
//      }
//      if (i == IDOK)  {         // take new configuration
//        if (ReInitTarget())  {  // failed...
//          StopTarget();         // final shutdown
//          PostMessage (hMfrm, Uv2Msg, MSG_UV2_TERMINATE, 0);
//        }
//      }
//    }
//  }
//  else  {                       // cancelled: restore previous config
//    MonConf = oCnf;
//  }
//}

/*
 * uVision2 wants to update all modeless dialogs.
 */

static void DlgUpdate()  {        // Update all modeless extension dialogs
  const DYMENU* pM = Menu;

  while (pM->nDelim != -1)  {             // while not end of Menu-list
    if (pM->pDlg && pM->pDlg->Update)  {  // if dialog has been created
      pM->pDlg->Update();                 // then call it's update function
    }
    ++pM;                                 // next menu entry.
  }
}


/*
 * Need to close all currently open modeless dialogs
 */

static void CloseAllDlg()  {
  const DYMENU* pM = Menu;

  while (pM->nDelim != -1)  {
    if (pM->pDlg && pM->pDlg->hw)  {        // Dialog is visible
      pM->pDlg->Kill(pM->pDlg);            // update contents
    }
    ++pM;
  }
}




//--- Interface functions between AGDI and target follow
//------------------------------------------------------

static HRESULT ReadBytes(MemorySegment seg, DWORD startAddress, BYTE* pB, DWORD len)
{
    SAFEARRAY*  psa = nullptr;
    HRESULT     hr;
    void*       pv;

    CAutoCriticalSection sect(&gs_cs);
    Debug(TEXT("GetMemory %u (Address: %08X, Number of Bytes to read: %u)\n"), GetCurrentThreadId(), startAddress, len);

    if(!gs_pProgrammer)
        return(E_UNEXPECTED);

    if(FAILED(hr = gs_pProgrammer->GetMemory(seg, static_cast<long>(startAddress & 0x00FFFFFF), static_cast<long>(len), nullptr, &psa)))
    {
        Debug(TEXT("ReadBytes FAILED: 0x%08X\n"), hr);
        return(hr);
    }

    if(SUCCEEDED(hr = SafeArrayAccessData(psa, &pv)))
    {
        CopyMemory(pB, pv, len);
        SafeArrayUnaccessData(psa);
    }

    SafeArrayDestroy(psa);
    return(hr);
}

/*
 * Read 'nMany' bytes out of data space into buffer 'pB'
 *  return: 0:=Ok, else 'ErrorAddress | (amDATA << 24)'
 */

static UL32 ReadData (BYTE *pB, DWORD nAdr, DWORD nMany)  {
    if (CacheValid (pB, nAdr, nMany))  {   // use cache, if possible
        return (0);
    }

//---TODO:
//---if Ok, then return 0, else error-address
//   adr |= (amDATA << 24);   // make address uVision2 conforming
    if(FAILED(ReadBytes(segData, nAdr, pB, nMany)))
        return(MAKE_UVISION_DATAPTR(nAdr));

    return (0);                 // say Ok.
}

/*
 * Read 'nMany' bytes out of the idata space into buffer 'pB'
 *  return: 0:=Ok, else 'ErrorAddress | (amIDATA << 24)'
 */

static UL32 ReadIdata (BYTE *pB, DWORD nAdr, DWORD nMany)  {
    if (CacheValid (pB, nAdr, nMany))  {   // use cache, if possible
        return (0);
    }
//---TODO:
//---if Ok, then return 0, else error-address
//   adr |= (amIDATA << 24);  // make address uVision2 conforming
    if(FAILED(ReadBytes(segIData, nAdr, pB, nMany)))
        return(MAKE_UVISION_IDATAPTR(nAdr));

    return (0);                 // say Ok.

}

/*
 * Read 'nMany' sfr values out of sfr-data space into buffer 'pB'
 *  return: 0:=Ok, else 'ErrorAddress | (amDATA << 24)'
 */

static UL32 ReadSFR (BYTE *pB, DWORD nAdr, DWORD nMany)  {
    return(ReadData(pB, MAKE_UVISION_DATAPTR(nAdr), nMany));
}

/*
 * Read 'nMany' bytes out of the xdata space into buffer 'pB'
 *  return: 0:=Ok, else 'ErrorAddress | (amXDATA << 24)'
 */

static UL32 ReadXdata (BYTE *pB, DWORD nAdr, DWORD nMany)  {
    if (CacheValid (pB, nAdr, nMany))  {   // use cache, if possible
        return (0);
    }
//---TODO:
//---if Ok, then return 0, else error-address
//   adr |= (amIDATA << 24);  // make address uVision2 conforming
    if(FAILED(ReadBytes(segXData, nAdr, pB, nMany)))
        return(MAKE_UVISION_XDATAPTR(nAdr));

    return (0);                 // say Ok.
}

/*
 * Read 'nMany' bytes out of the code space into buffer 'pB'
 *  return: 0:=Ok, else 'ErrorAddress | (amCODE << 24)'
 */

static UL32 ReadCode (BYTE *pB, DWORD nAdr, DWORD nMany)  {
    if (CacheValid (pB, nAdr, nMany))  {   // use cache, if possible
        return (0);
    }
    else  {
        memset (pB, 0, nMany);    // for now, we return 0 bytes (NOP's)
        // txtout("Reading %u Bytes of Code Memory beginning at 0x%04X.\n", nMany, nAdr);
        if(FAILED(ReadBytes(segCode, nAdr, pB, nMany)))
            return(MAKE_UVISION_CODEPTR(nAdr));
    }

    //---TODO:
    //---if Ok, then return 0, else error-address
    //   adr |= MAKE_UVISION_CODEPTR(0);   // make address uVision2 conforming
    Debug(TEXT("ReadCode succeeded:\n"));
    return (0);                 // say Ok.
}

static HRESULT WriteBytes(MemorySegment seg, DWORD startAddress, const BYTE* pData, DWORD len)
{
    SAFEARRAYBOUND  sab;
    SAFEARRAY*      psa;
    HRESULT         hr;
    void*           pv;

    Debug(TEXT("WriteBytes %u\n"), GetCurrentThreadId());

    sab.lLbound = 0;
    sab.cElements = len;

    if(nullptr == (psa = SafeArrayCreate(VT_UI1, 1, &sab)))
        return(E_OUTOFMEMORY);

    if(SUCCEEDED(hr = SafeArrayAccessData(psa, &pv)))
    {
        CopyMemory(pv, pData, len);
        hr = gs_pProgrammer->SetMemory(seg, static_cast<long>(startAddress) & 0x00FFFFFFL, &psa, VARIANT_FALSE, nullptr);
        SafeArrayUnaccessData(psa);
    }

    SafeArrayDestroy(psa);
    return(hr);
}

/*
 * Write 'nMany' bytes of code into target
 *  return: 0:=Ok, else 'ErrorAddress | (amCODE << 24)'
 */

static UL32 WriteCode (const BYTE* pB, DWORD nAdr, DWORD nMany)  {
//---TODO:
//---if Ok, then return 0, else error-address
//   adr |= (amCODE << 24);   // make address uVision2 conforming
    // txtout("Writing %u Bytes of Code Memory beginning at 0x%08X\n", nMany, nAdr);

    UNREFERENCED_PARAMETER(pB);
    UNREFERENCED_PARAMETER(nAdr);
    UNREFERENCED_PARAMETER(nMany);
    Debug(TEXT("WriteCode %u\n"), GetCurrentThreadId());

//    if(FAILED(WriteBytes(segCode, nAdr, pB, nMany)))
//        return(MAKE_UVISION_CODEPTR(nAdr));

    return (0);                 // say Ok.
}

/*
 * Write 'nMany' bytes of xdata into target
 *  return: 0:=Ok, else 'ErrorAddress | (amXDATA << 24)'
 */

static UL32 WriteXdata (const BYTE*  pB, DWORD nAdr, DWORD nMany)  {
//---TODO:
//---if Ok, then return 0, else error-address
//   adr |= (amXDATA << 24);  // make address uVision2 conforming
    if(FAILED(WriteBytes(segXData, nAdr, pB, nMany)))
        return(MAKE_UVISION_XDATAPTR(nAdr));

    return (0);                 // say Ok.
}

/*
 * Write 'nMany' bytes of idata into target
 *  return: 0:=Ok, else 'ErrorAddress | (amIDATA << 24)'
 */

static UL32 WriteIdata (const BYTE* pB, DWORD nAdr, DWORD nMany)  {
//---TODO:
//---if Ok, then return 0, else error-address
//   adr |= (amIDATA << 24);  // make address uVision2 conforming
    if(FAILED(WriteBytes(segIData, nAdr, pB, nMany)))
        return(MAKE_UVISION_IDATAPTR(nAdr));

    return (0);                 // say Ok.
}

/*
 * Write 'nMany' bytes of data into target
 *  return: 0:=Ok, else 'ErrorAddress | (amDATA << 24)'
 */

static UL32 WriteData (const BYTE* pB, DWORD nAdr, DWORD nMany)  {
//---TODO:
//---if Ok, then return 0, else error-address
//   adr |= (amDATA << 24);   // make address uVision2 conforming
    if(FAILED(WriteBytes(segData, nAdr, pB, nMany)))
        return(MAKE_UVISION_DATAPTR(nAdr));

    return (0);                 // say Ok.
}

/*
 * Write 'nMany' sfr's to target
 *  return: 0:=Ok, else 'ErrorAddress | (amDATA << 24)'
 */

static UL32 WriteSFR (const BYTE* pB, DWORD nAdr, DWORD nMany)  {
//---TODO:
//---if Ok, then return 0, else error-address
//   adr |= (amDATA << 24);   // make address uVision2 conforming
    return(WriteData(pB, nAdr, nMany));
}

static HRESULT String2ChipId(BSTR bstrId, ChipId* pid)
{
    ITypeInfo* pTypeInfo = nullptr;
    TYPEATTR* pAttr = nullptr;
    UINT cnt;
    HRESULT hr;
    bool blFound = false;

    if(SUCCEEDED(hr = ::GetTypeInfoOfChipIdEnumeration(gs_pProgrammer, &pTypeInfo)))
    {
        if(SUCCEEDED(hr = pTypeInfo->GetTypeAttr(&pAttr)))
        {
            if(pAttr->typekind == TKIND_ENUM)
            {
                VARIANT var;
                VariantInit(&var);

                for(UINT idx = 0; (idx < pAttr->cVars) && (!blFound); ++idx)
                {
                    VARDESC* pDesc = nullptr;

                    if(SUCCEEDED(pTypeInfo->GetVarDesc(idx, &pDesc)))
                    {
                        BSTR bstr = nullptr;

                        if(SUCCEEDED(pTypeInfo->GetNames(pDesc->memid, &bstr, 1, &cnt)))
                        {
                            if(VARCMP_EQ == VarBstrCmp(bstr, bstrId, CThreadLocale::getLCID(), NORM_IGNORECASE))
                            {
                                if(SUCCEEDED(VariantChangeTypeEx(&var, pDesc->lpvarValue, CThreadLocale::getLCID(), 0, VT_I4)))
                                    *pid = static_cast<ChipId>(V_I4(&var));
                                else
                                    *pid = UnknownChip;

                                blFound = true;
                            }
                            SysFreeString(bstr);
                        }
                        pTypeInfo->ReleaseVarDesc(pDesc);
                        VariantClear(&var);
                    }
                }
            }
            pTypeInfo->ReleaseTypeAttr(pAttr);
        }
        pTypeInfo->Release();
    }

    if(SUCCEEDED(hr))
    {
        if(!blFound)
        {
            txtout("Targets of Type \"%ls\" are not supported by this debug adapter.\n", bstrId);
            hr = E_FAIL;
        }
    }

    return(hr);
}

static HRESULT ChipId2String(ChipId id, BSTR* pbstr)
{
    ITypeInfo* pTypeInfo = nullptr;
    TYPEATTR* pAttr = nullptr;
    UINT cnt;
    HRESULT hr;

    if(SUCCEEDED(hr = ::GetTypeInfoOfChipIdEnumeration(gs_pProgrammer, &pTypeInfo)))
    {
        if(SUCCEEDED(hr = pTypeInfo->GetTypeAttr(&pAttr)))
        {
            if(pAttr->typekind == TKIND_ENUM)
            {
                bool blDone = false;
                VARIANT var;
                VariantInit(&var);

                for(UINT idx = 0; (!blDone) && (idx < pAttr->cVars) && SUCCEEDED(hr); ++idx)
                {
                    VARDESC* pDesc = nullptr;

                    if(SUCCEEDED(hr = pTypeInfo->GetVarDesc(idx, &pDesc)))
                    {
                        if(SUCCEEDED(hr = VariantChangeTypeEx(&var, pDesc->lpvarValue, CThreadLocale::getLCID(), 0, VT_I4)))
                        {
                            if(V_I4(&var) == static_cast<int>(id))
                            {
                                hr = pTypeInfo->GetNames(pDesc->memid, pbstr, 1, &cnt);
                                blDone = true;
                            }
                        }
                        pTypeInfo->ReleaseVarDesc(pDesc);
                        VariantClear(&var);
                    }
                }
            }
            else
            {
                hr = E_FAIL;
            }
            pTypeInfo->ReleaseTypeAttr(pAttr);
        }
        pTypeInfo->Release();
    }

    return(hr);
}

static HRESULT VerifyTargetChipId()
{
    HRESULT hr;
    ChipId idAttached = UnknownChip;
    ChipId idWanted;
    char* pContext = nullptr;
    char* ptmp;
    BSTR bstrWanted;
    BSTR bstrAttached;
    long lSupplyVoltage = 0;

    gs_pProgrammer->get_TargetVoltage(&lSupplyVoltage);

    if(lSupplyVoltage < 1800)
    {
        CThreadLocale loc(MAKELCID(MAKELANGID(PRIMARYLANGID(CThreadLocale::getLANGID()), SUBLANG_DEFAULT), SORT_DEFAULT));
        ptmp = nullptr;
        FormatMessageA(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_HMODULE | FORMAT_MESSAGE_IGNORE_INSERTS, hInst, static_cast<DWORD>(SMARTICE04_E_NOTARGET), 0, reinterpret_cast<char*>(&ptmp), 0, nullptr);

        if(ptmp)
        {
            txtout_buildwnd(ptmp);
            LocalFree(ptmp);
        }

        return(SMARTICE04_E_NOTARGET);
    }

    if(FAILED(hr = gs_pProgrammer->get_ChipId(&idAttached)))
    {
        txtout_buildwnd("Could not read the target chip id (0x%08X)\n", hr);
        return(hr);
    }

    hr = E_UNEXPECTED;
    ptmp = _strdup(pdbg->PeriArgs);

    for(char* pToken = strtok_s(((ptmp) ? ptmp : pdbg->PeriArgs), "-/", &pContext); pToken; pToken = strtok_s(nullptr, "-/", &pContext))
    {
        if(*pToken++ == 'p')
        {
            TrimA(pToken);

            if(SUCCEEDED(hr = BSTRFromStringA(pToken, &bstrWanted)))
            {
                if(SUCCEEDED(hr = String2ChipId(bstrWanted, &idWanted)))
                {
                    if(idWanted != idAttached)
                    {
                        txtout_buildwnd("Wrong chip attached to the debug adapter.\n");
                        hr = E_FAIL;
                        bstrAttached = nullptr;

                        if(SUCCEEDED(ChipId2String(idAttached, &bstrAttached)))
                        {
                            txtout_buildwnd("You wish to debug a %ls, attached is a %ls.\n", bstrWanted, ((bstrAttached) ? bstrAttached : L"unknown chip"));
                            SysFreeString(bstrAttached);
                        }
                    }
                }
                SysFreeString(bstrWanted);
            }
            else
            {
                hr = E_OUTOFMEMORY;
            }
        }
    }

    free(ptmp);
    return(hr);
}

/*
 * Initialize your target communication
 * Return 0 if Ok, 1 if failed.
 */

static U32 InitTarget()  {
    ISmartICE04Devices* pDevices = nullptr;
    VARIANT             varIndex, var;
    U32                 ret = 1;

    Debug(TEXT("InitTarget %u\n"), GetCurrentThreadId());

    if(gs_pProgrammer)
        return(0);

    VariantInit(&varIndex);
    VariantInit(&var);

    __try
    {
        HRESULT hr = S_OK;

        if(nullptr == (pDevices = new (std::nothrow) CSmartICE04Devices(&hr)))
        {
            hr = E_OUTOFMEMORY;
            __leave;
        }

        if(FAILED(hr))
            __leave;

        if(FAILED(hr = GetDebugAdapterId(pDevices, &V_BSTR(&varIndex))))
            __leave;

        V_VT(&varIndex) = VT_BSTR;

        if(hr != S_OK)
        {
            hr = __HRESULT_FROM_WIN32(ERROR_CANCELLED);
            __leave;
        }

        if(FAILED(hr = pDevices->get_Item(varIndex, &var)))
            __leave;

        if(FAILED(hr = VariantChangeTypeEx(&var, &var, CThreadLocale::getLCID(), 0, VT_DISPATCH)))
            __leave;

        if(!V_DISPATCH(&var))
        {
            hr = E_UNEXPECTED;
            __leave;
        }

        if(!InterlockedCompareExchange(&gs_blInitialized, -1, 0))
            InitializeCriticalSection(&gs_cs);

        if(FAILED(hr = V_DISPATCH(&var)->QueryInterface(IID_ISmartICE04, reinterpret_cast<void**>(&gs_pProgrammer))))
            __leave;

        if(FAILED(hr = VerifyTargetChipId()))
            __leave;

        if(SUCCEEDED(hr = gs_pProgrammer->EnterDebugMode(VARIANT_TRUE)))
            ret = 0;
    }
    __finally
    {
        VariantClear(&var);
        VariantClear(&varIndex);

        if(pDevices)
            pDevices->Release();

        Debug(TEXT("InitTarget %s, %u\n"), (!ret) ? TEXT("succeeded") : TEXT("failed"), GetCurrentThreadId());
    }

    return(ret);
}


/*
 * Target settings have changed (Baudrate or ComPort for example)
 * Return 0 if Ok, 1 if failed.
 */
//static U32 ReInitTarget()  {
//    //---TODO: shutdown target
//    //---      reinit communication with current 'MonConf' values
//
//    if(gs_pProgrammer)
//    {
//        StopTarget();
//        InitTarget();
//    }
//
//    return (0);            // say 'Ok.'
//}


/*
 * Stop your target communication
 * Free all resources (dynamic memory, ...)
 */

static void StopTarget()  {
    static volatile LONG s_blIsStopping = FALSE;
    DWORD dwTickCount = GetTickCount();
    Debug(TEXT("StopTarget %u, gs_pProgrammer = 0x%08X\n"), GetCurrentThreadId(), gs_pProgrammer);

    if(InterlockedCompareExchange(&s_blIsStopping, -1, 0))
        return;

    if(gs_pProgrammer)
    {
        CAutoCriticalSection sect(&gs_cs);
        gs_pProgrammer->Halt();

        while(gs_Running)
        {
            if((GetTickCount() - dwTickCount) > 10000)
                break;

            SwitchToThread();
        }

        gs_pProgrammer->EnterDebugMode(VARIANT_FALSE);

        if(gs_Running)
        {
            Debug(TEXT("Still Running\n"));
            dwTickCount = GetTickCount();

            while(gs_Running)
            {
                if((GetTickCount() - dwTickCount) > 10000)
                {
                    InterlockedExchange(&gs_Running, FALSE);
                    break;
                }
                SwitchToThread();
            }
        }

        gs_pProgrammer->Release();
        gs_pProgrammer = nullptr;
    }

    if(InterlockedCompareExchange(&gs_blInitialized, 0, -1))
        DeleteCriticalSection(&gs_cs);

    FreeCache();        // free allocated cache memory.
    InterlockedExchange(&s_blIsStopping, 0);
}

/*
 * Reset your target. Set PC to whatever required value
 */

static void ResetTarget()  {
    Debug(TEXT("ResetTarget %u\n"), GetCurrentThreadId());
    if(gs_pProgrammer)
    {
        gs_pProgrammer->EnterDebugMode(VARIANT_FALSE);
        gs_pProgrammer->EnterDebugMode(VARIANT_TRUE);
    }
    REG51.nPC = MAKE_UVISION_CODEPTR(0);
    ZeroMemory(gs_Breakpoints, sizeof(gs_Breakpoints));
}


/*
 * Read PC out of target
 */


static UL32 ReadPC()  {
    CAutoCriticalSection sect(&gs_cs);
    UL32    nPC = MAKE_UVISION_CODEPTR(REG51.nPC);
    long    pc;

    if(gs_pProgrammer)
    {
        if(SUCCEEDED(gs_pProgrammer->GetPc(&pc)))
        {
            nPC = MAKE_UVISION_CODEPTR(pc);
            Debug(TEXT("ReadPC SUCCEEDED\n"));
        }
    }

    Debug(TEXT("ReadPC = 0x%08X, %u\n"), nPC, GetCurrentThreadId());
    return(nPC);
}


/*
 * Write PC out to target
 */

static void WritePC (UL32 nPC)  {
//---TODO: write PC to target
    Debug(TEXT("WritePC to 0x%04X, %u\n"), (nPC & 0x00FFFFFF), GetCurrentThreadId());

    if(gs_pProgrammer)
    {
        if(SUCCEEDED(gs_pProgrammer->DebugInstr(MAKE_INSTRUCTION3(0x02, HIBYTE(nPC), LOBYTE(nPC)), nullptr)))  // ljmp nPC
            REG51.nPC = nPC;                // keep in shadow registers
    }
}



/*
 * Read all registers out of target and fill the REG51 structure
 */

static void GetRegs()  {
Debug(TEXT("GetRegs %u\n"), GetCurrentThreadId());
    BYTE bSFR[sizeof(REG51.sfr)];

    if (RegUpToDate) return;        // already up to date
    if (iRun) return;               // currently executing, can't access
    curPC = REG51.nPC = ReadPC();
//---TODO: fetch regs from target and put them into REG51
//---if connection is broken for some reason, then set PlayDead:=1 !

    if(ReadSFR(bSFR, 0x80, sizeof(bSFR)))
    {
        Debug(TEXT("GetRegs FAILED!\n"));
        PlayDead = 1;
    }
    else
    {
        CopyMemory(REG51.sfr, bSFR, sizeof(bSFR));

        if(ReadData(REG51.Rn, MAKE_UVISION_DATAPTR(static_cast<DWORD>(REG51.sfr[REGADDRESS2INDEX(PSW)] & 0x18)), sizeof(REG51.Rn)))
            PlayDead = 1;
    }
}


/*
 * Write all registers to target
 */

static void SetRegs (const RG51* pR)  {
    Debug(TEXT("SetRegs %u\n"), GetCurrentThreadId());
    //--- write pR->Rn[0] ... pR->Rn[7]
    //--- write pR->sp, pR->psw, pR->b, pR->acc, pR->dpl, pR->dph, pR->nPC
    WritePC (pR->nPC);
    //---if connection is broken for some reason, then set PlayDead:=1 !

    if(WriteSFR(pR->sfr, 0x80, sizeof(pR->sfr)) || WriteData(pR->Rn, static_cast<DWORD>(pR->sfr[REGADDRESS2INDEX(PSW)] & 3), sizeof(pR->Rn)))
        PlayDead = 1;
}


/*
 * Stop execution of user program
 */

static U32 StopExec()  {
Debug(TEXT("StopExec %u\n"), GetCurrentThreadId());
    if (iRun)  {                  // if currently executing
        //---TODO: Stop exec
        //---if successful, return (1), else (0)
        if(!gs_pProgrammer)
            return(0);

        if(FAILED(gs_pProgrammer->Halt()))
            return(0);
    }
    return (1);                   // 'Stopped'
}

/*
 * Invalidate everything which may be invalid after Go or Step
 *  (Registers, Caches, etc.)
 */

static void Invalidate()  {
  RegUpToDate = 0;                       // invalidate Registers
  ClearCaR(MAKE_UVISION_DATAPTR(0x00), 0);  // invalidate data-cache
  ClearCaR(MAKE_UVISION_IDATAPTR(0x80), 128);// invalidate idata-cache
  ClearCaR(MAKE_UVISION_XDATAPTR(0x0000), 0x10000);// invalidate xdata-cache
}


/*
 * Execute a single instruction, return 1 if ok, else 0
 */

static UL32 Step()  {

    CAutoCriticalSection sect(&gs_cs);
    Debug(TEXT("Step %u\n"), GetCurrentThreadId());

    if(!gs_pProgrammer)
        return(0);

    if(FAILED(gs_pProgrammer->StepInstr(nullptr)))
        return(0);

    return (1);
}


/*
 * Start execution.  Stop when a Bp is reached
 */


static void GoCmd()  {
Debug(TEXT("GoCmd %u\n"), GetCurrentThreadId());
    VARIANT_BOOL    vblIsInDebugMode = VARIANT_TRUE;
    HRESULT         hr;
    long            status = 0;

    if(!gs_pProgrammer)
        return;

    InterlockedIncrement(&gs_Running);

    if(FAILED(hr = gs_pProgrammer->Resume()))
    {
        txtout("FAILED to start cpu (0x%08X)\n", hr);
        PlayDead = 1;
        InterlockedExchange(&gs_Running, FALSE);
        return;
    }

    while(!(status & static_cast<long>(CpuHalted)))
    {
        Sleep(100);

        if(!gs_pProgrammer)
            break;

        if(FAILED(hr = gs_pProgrammer->ReadStatus(&status)))
        {
            if(hr == SMARTICE04_E_NOTINPGMMODE)
            {
                vblIsInDebugMode = VARIANT_FALSE;
            }
            else if(FAILED(hr = gs_pProgrammer->ReadStatus(&status)))
            {
                txtout("FAILED to read cpu status (0x%08X)\n", hr);
                PlayDead = 1;
                break;
            }
        }
        else if(FAILED(hr = gs_pProgrammer->get_IsInDebugMode(&vblIsInDebugMode)))
        {
            txtout("Debug adapter lost.\n");
            PlayDead = 1;
            break;
        }

        if(!vblIsInDebugMode || PlayDead)
        {
            if(SUCCEEDED(hr) && PlayDead)
            {
                hr = E_UNEXPECTED;
                StopTarget();
            }

            txtout("Device unintentionally lost.\n");
            PlayDead = 1;
            break;
        }
    }

    if(SUCCEEDED(hr) && vblIsInDebugMode)
    {
        if(status & static_cast<long>(HaltStatus))
        {
            const UL32 pc = ReadPC() - 1;

            for(uintptr_t x = 0; x < _countof(gs_Breakpoints); ++x)
            {
                if(pc == gs_Breakpoints[x].address)
                    WritePC(pc);
            }
        }
    }

    InterlockedDecrement(&gs_Running);
    Debug(TEXT("Execution stopped, status = 0x%X, %u\n"), status, GetCurrentThreadId());
}


/*
 * Set/Clear Breakpoint at address pB->Adr
 */

static int SetClrBp (int set, const AG_Bps* pB)  {
    unsigned long x;
    int ret = 0;

//    txtout("%s Breakpoint at 0x%04X\n", ((set) ? "Set" : "Clear"),  (pB->Adr & 0x00FFFFFF));
//    Debug(TEXT("%s Breakpoint at 0x%04X, %u\n"), ((set) ? TEXT("Set") : TEXT("Clear")),  (pB->Adr & 0x00FFFFFF), GetCurrentThreadId());

    if(!gs_pProgrammer)
        return(0);

    InterlockedIncrement(&gs_Running);

    if (set)  {                   // set address break
        //---TODO: give target a message about Set-Bp at address 'pB->Adr'
        //---      if ok, then return (1), else (0)
        //  if (failed)  {
        //    txtout ((char *) szNBP, pB->Adr);
        //    nR = 0;
        //  }

        for(x = 0; x < _countof(gs_Breakpoints); ++x)
        {
            if((gs_Breakpoints[x].address == pB->Adr) && (gs_Breakpoints[x].enabled))
            {
                ret = 1;
                break;
            }
        }

        if(x >= _countof(gs_Breakpoints))
        {
            for(x = 0; x < _countof(gs_Breakpoints); ++x)
            {
                if(!gs_Breakpoints[x].enabled)
                {
                    if(FAILED(gs_pProgrammer->SetHwBreakpoint(static_cast<long>(x), VARIANT_TRUE, static_cast<long>(pB->Adr & 0x00FFFFFF))))
                    {
                        txtout("FAILED to set breakpoint at address 0x%08X\n", pB->Adr);
                    }
                    else
                    {
                        gs_Breakpoints[x].address = pB->Adr;
                        gs_Breakpoints[x].enabled = TRUE;
                        ret = 1;
                    }
                    break;
                }
            }

            if(x >= _countof(gs_Breakpoints))
            {
                txtout("Maximum number of breakpoints reached (%u)\n", x);
            }
        }
    }
    else  {                       // clear address break
        //---TODO: give target a message about Clear-Bp at address 'pB->Adr'
        //---      if ok, then return (1), else (0)
        for(x = 0; x < _countof(gs_Breakpoints); ++x)
        {
            if((pB->Adr == static_cast<UL32>(gs_Breakpoints[x].address)) && (gs_Breakpoints[x].enabled))
            {
                if(SUCCEEDED(gs_pProgrammer->SetHwBreakpoint(static_cast<long>(x), VARIANT_FALSE, 0xFFFF)))
                {
                    gs_Breakpoints[x].enabled = FALSE;
                    ret = 1;
                    break;
                }
            }
        }
    }

#ifdef _DEBUG
    unsigned cnt = 0;

    for(x = 0; x < _countof(gs_Breakpoints); ++x)
        if(gs_Breakpoints[x].enabled)
            ++cnt;

    Debug(TEXT("There are currently %u breakpoints enabled\n"), cnt);
#endif

    InterlockedDecrement(&gs_Running);
    return(ret);
}


/*
 * Send text Error message to Build window
 */

static void PgmError (const char *pCmd)  {
  txtout ("*** Error: %s\n", pCmd);
}

static void PgmError(const wchar_t* pCmd)
{
    int nLen = WideCharToMultiByte(CP_ACP, WC_NO_BEST_FIT_CHARS, pCmd, -1, nullptr, 0, nullptr, nullptr);

    if(!nLen)
        return;

    __try
    {
        char* psz = static_cast<char*>(_malloca(static_cast<unsigned>(nLen) * sizeof(*psz)));

        if(psz)
        {
            if(WideCharToMultiByte(CP_ACP, WC_NO_BEST_FIT_CHARS, pCmd, -1, psz, nLen, nullptr, nullptr))
            {
                while(nLen--)
                {
                    if((psz[nLen] == '\r') || (psz[nLen] == '\n'))
                        psz[nLen] = ' ';
                }

                PgmError(psz);
            }

            _freea(psz);
        }
    }
    __except((GetExceptionCode() == EXCEPTION_STACK_OVERFLOW) ? EXCEPTION_EXECUTE_HANDLER : EXCEPTION_CONTINUE_SEARCH)
    {
        _resetstkoflw();
    }
}

static void PgmError(HRESULT hr)
{
    HRESULT tmp;

    if(HRESULT_FACILITY(hr) == FACILITY_ITF)
    {
        IErrorInfo* pInfo = nullptr;

        if(S_OK == GetErrorInfo(0, &pInfo))
        {
            BSTR bstr = nullptr;
            if(SUCCEEDED(tmp = pInfo->GetDescription(&bstr)))
            {
                PgmError(bstr);
                SysFreeString(bstr);
            }
            pInfo->Release();
        }
        else
        {
            tmp = E_FAIL;
        }
    }
    else
    {
        PTSTR pBuffer = nullptr;
        if(!FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_FROM_SYSTEM, nullptr, static_cast<DWORD>(hr), CThreadLocale::getLANGID(), reinterpret_cast<LPTSTR>(&pBuffer), 0, nullptr))
        {
            tmp = E_FAIL;
        }
        else
        {
            tmp = S_OK;
            PgmError(pBuffer);
        }
        LocalFree(pBuffer);
    }

    if(FAILED(tmp))
    {
        char sz[32];
        StringCchPrintfA(sz, _countof(sz), "FAILURE: %ld (0x%08lX)", hr, hr);
        PgmError(sz);
    }
}

//--- Flash Download interface functions follow
//------------------------------------------------------

/*
 * Initialize Flash programmer, download programmer firmware.
 * Return 0 if Ok, 1 if failed.
 */

static int InitProgrammer()
{
    Debug(TEXT("InitProgrammer %u\n"), GetCurrentThreadId());
    FlashProg = 1;
    return (0);            // say 'Ok.'
}


/*
 * Stop Flash programmer, disconnect device power
 * Return 0 if Ok, 1 if failed.
 */

static int ExitProgrammer()
{
    FlashProg = 0;
    Debug(TEXT("ExitProgrammer %u\n"), GetCurrentThreadId());
    return (0);            // say 'Ok.'
}


/*
 * Enter programming mode for the flash device
 * Return 0 if Ok, 1 if failed.
 */

static int EnterProgMode()
{
    HRESULT hr;
    Debug(TEXT("EnterProgMode %u\n"), GetCurrentThreadId());

    if(!gs_pProgrammer)
        return(1);

    if(FAILED(hr = gs_pProgrammer->EnterDebugMode(VARIANT_TRUE)))
    {
        PgmError(hr);
        return(1);
    }

    return (0);            // say 'Ok.'
}


/*
 * Erase the flash device
 * Return 0 if Ok, 1 if failed.
 */

static int EraseFlash()
{
Debug(TEXT("EraseFlash %u\n"), GetCurrentThreadId());
    HRESULT hr;

    if(!gs_pProgrammer)
        return(1);

    if(FAILED(hr = gs_pProgrammer->ChipErase()))
    {
        PgmError(hr);
        return(1);
    }

    return (0);            // say 'Ok.'
}


/*
 * Write data to Flash device
 * Return 0 if Ok, 1 if failed.
 */

static int WriteToFlash (DWORD nAdr, const BYTE * pb, DWORD nCnt, DWORD *eAdr, CAgdiProgress* pProgress)
{
    SAFEARRAYBOUND  sab;
    SAFEARRAY*      psa;
    VARIANT         varProgress;
    HRESULT         hr;
    void*           pData = nullptr;

//---TODO: Add Flash Write Code here
//---Note: nAdr has uVision2 address convention
//---if write fails, set write error address and return 1
    *eAdr = nAdr;

    if((!gs_pProgrammer) || (((nAdr >> 24) & 0xFF) != amCODE))
        return(1);

    sab.lLbound = 0;
    sab.cElements = nCnt;

    if(nullptr == (psa = SafeArrayCreate(VT_UI1, 1, &sab)))
    {
        PgmError(E_OUTOFMEMORY);
        return(1);
    }

    if(SUCCEEDED(hr = SafeArrayAccessData(psa, &pData)))
    {
        CopyMemory(pData, pb, nCnt);
        SafeArrayUnaccessData(psa);
        VariantInit(&varProgress);
        V_VT(&varProgress) = VT_UNKNOWN;
        V_UNKNOWN(&varProgress) = pProgress;
        V_UNKNOWN(&varProgress)->AddRef();
        hr = gs_pProgrammer->SetMemory(segCode, static_cast<long>(nAdr & 0x00FFFFFF), &psa, VARIANT_FALSE, &varProgress);
        VariantClear(&varProgress);
    }

    SafeArrayDestroy(psa);

    if(FAILED(hr))
    {
        PgmError(hr);
        return(1);
    }

    return (0);            // say 'Ok.'
}


/*
 * Verify the content of Flash device
 * Return 0 if Ok, 1 if failed.
 */

static int VerifyFlash (DWORD nAdr, const BYTE * pb, DWORD nCnt, DWORD *eAdr, CAgdiProgress* pProgress)
{
    SAFEARRAY*  psa = nullptr;
    VARIANT     varProgress;
    HRESULT     hr;
    BYTE*       pData = nullptr;

//---TODO: Verify Flash Code here
//---Note: nAdr has uVision2 address convention
//---if verify fails, set verify error address and return 1
    *eAdr = nAdr;

    if((!gs_pProgrammer) || (((nAdr >> 24) & 0xFF) != amCODE))
        return(1);

    VariantInit(&varProgress);
    V_VT(&varProgress) = VT_UNKNOWN;
    V_UNKNOWN(&varProgress) = pProgress;

    if(FAILED(hr = gs_pProgrammer->GetMemory(segCode, static_cast<long>(nAdr & 0x00FFFFFF), static_cast<long>(nCnt), &varProgress, &psa)))
    {
        PgmError(hr);
        return(1);
    }

    if(SUCCEEDED(hr = SafeArrayAccessData(psa, reinterpret_cast<void**>(&pData))))
    {
        nAdr = 0;
        while(nCnt)
        {
            --nCnt;
            if(*pb++ != *pData++)
            {
                *eAdr += nAdr;
                break;
            }
            ++nAdr;
        }
    }

    SafeArrayDestroy(psa);

    if(FAILED(hr))
    {
        PgmError(hr);
        return(1);
    }

    return((!nCnt) ? 0 : 1);
}

/*
 * Initialize target Flash device.
 * Return 0 if Ok, 1 if failed.
 */

static int InitDevFlash()
{
//---We use 'EnterProgMode()' function to activate the programming mode of the device
//---when a special procedure to activate the programming mode is required.
//---If this is not required, we can skip this function.
  int nE = EnterProgMode ();
  if (nE) {
    PgmError ("Programming failed. Device in programming socket did not respond.");
    return (1);                         // failed
  }

  if (!(MonConf.Opt & FLASH_ERASE)) {   // check if Download Function-Erase Flash enabled
    return (0);
  }

//---We erase the target flash here and prepare for flash writing.
  nE = EraseFlash ();
  if (nE) {
    PgmError ("Flash Erase failed.");
    return (1);                         // failed Flash not Erased
  }
  txtout ("Flash Erase Done.\n");
  return (0);                           // Ok.
}


/*
 * Write data for Flash Memory to Target
 *
 * If target hardware has installed SRAM that is big enough to hold the
 * complete Flash image data, it is a good practice to use it as a buffer.
 * This would speedup the flash verify process a lot if a big flash with
 * big amount of data is programmed using 'slow RS232' interface.
 */

static int FlashLoad()
{
    FLASHPARM  *pF;
    DWORD  bCnt,errAdr;
    char szMsg[80];
    int nE = 0;

    Debug(TEXT("FlashLoad %u\n"), GetCurrentThreadId());

    if(MonConf.Opt & FLASH_PROGRAM)
    {
        //---Write Flash memory if enabled
        InitProgress("Loading...");
        //---Get Flash Parameters from S8051.DLL
        pF = static_cast<FLASHPARM*>(UIntToPtr(pCbFunc(AG_CB_GETFLASHPARAM, nullptr)));
        bCnt = 0;
        while (pF->many != 0) {                      // while more chunks..
            CAgdiProgress progress(static_cast<long>(bCnt), static_cast<long>(pF->ActSize), pCbFunc);
            nE = WriteToFlash(pF->start, pF->image, pF->many, &errAdr, &progress);
            if (nE) {                                // failed, send message
                StringCchPrintfA(szMsg, _countof(szMsg), "Flash Write Failed at %04lX", errAdr);
                PgmError(szMsg);
                break;
            }
            bCnt += pF->many;
            SetProgress(bCnt*100/pF->ActSize);
            pF = static_cast<FLASHPARM*>(UIntToPtr(pCbFunc (AG_CB_GETFLASHPARAM, pF)));// Get next parameters
        }

        if(!nE)
        {
            SetProgress(100);
            txtout("Flash Write Done: %d bytes programmed.\n", bCnt);
        }

        StopProgress();
    }

    if((MonConf.Opt & FLASH_VERIFY) && (!nE))
    {
        //---Verify Flash memory if enabled
        InitProgress("Verifying...");
        //---Get Flash Parameters from S8051.DLL
        pF = static_cast<FLASHPARM*>(UIntToPtr(pCbFunc(AG_CB_GETFLASHPARAM, nullptr)));
        bCnt = 0;
        while (pF->many != 0)  {                      // while more chunks..
            CAgdiProgress progress(static_cast<long>(bCnt), static_cast<long>(pF->ActSize), pCbFunc);
            nE = VerifyFlash (pF->start, pF->image, pF->many, &errAdr, &progress);
            if (nE) {                                   // failed, send message
                StringCchPrintfA(szMsg, _countof(szMsg), "Flash Verify Failed at %04lX", errAdr);
                PgmError(szMsg);
                break;
            }
            bCnt += pF->many;
            SetProgress (bCnt*100/pF->ActSize);
            pF = static_cast<FLASHPARM*>(UIntToPtr(pCbFunc (AG_CB_GETFLASHPARAM, pF)));// Get next parameters
        }

        if(!nE)
        {
            SetProgress(100);
            txtout("Flash Verify Done: %d bytes verified.\n", bCnt);
        }
    }

    StopProgress();
    ExitProgrammer();
    return(nE);
}

/*
 *----------- AGDI Basic Functions -----------------
 */


/*
 * AGDI-Init Function
 */

EXTERN_C U32 AGDI_DLLEXPORT AG_Init (U16 nCode, void *vp)  {
  static char execmd[] = "U $\n";
  U32     nE;

  nE = 0;
  switch (nCode & 0xFF00)  {
    case AG_INITFEATURES:     // Initialize & start the target
      PlayDead = 0;           // clear some variables...

//---We don't have special features here, so all variables are cleared.
//---uVision2 will query these features lateron.
//---Note: the 'supp' structure is defined in 'AGDI.H'

      supp.MemAccR  = 0;      // memory-access while running
      supp.RegAccR  = 0;      // register-access while running
      supp.hTrace   = 0;      // trace support
      supp.hCover   = 0;      // code coverage support
      supp.hPaLyze  = 0;      // Performance-Analyzer support
      supp.hMemMap  = 0;      // Memory-Map support
      supp.ResetR   = 0;      // Reset while running support

//---Note: if InitTarget() fails, then set nE=1, else nE=0.
//         if 1 is returned, then uVision will cancel using this driver

      FlashProg = 0;
      InitMem();                // init Cache memory

      nE = InitTarget();        // Initialize your target...

      if (nE == 0)  {           // initialization was Ok.
        InitRegs();             // define register layout for RegWindow
        pCbFunc (AG_CB_EXECCMD, execmd);  // force disassemble @curPC
      }

      NumRecs = 0;              // clear 'number of trace records'
      break;

    case AG_INITITEM:           // init item
      switch (nCode & 0x00FF)  {
        case AG_INITMENU:       // init extension menu
          *static_cast<DYMENU**>(vp) = Menu;
          break;
        case AG_INITEXTDLGUPD:    // init modeless extension dlg update function
          *static_cast<void**>(vp) = DlgUpdate;
          break;
        case AG_INITMHANDLEP:     // setup ptr to HWND of active modeless dlg
          pHwnd = static_cast<HWND*>(vp);
          break;
        case AG_INITPHANDLEP:     // pointer to parent handle (MainFrame)
          hMfrm = static_cast<HWND>(vp);
          break;
        case AG_INITINSTHANDLE:   // pointer to Agdi-instance handle
          hInst = static_cast<HMODULE>(vp);
          break;
        case AG_INITBPHEAD:       // pointer to head of Bp-List
          pBhead = static_cast<AG_Bps**>(vp);
          break;
        case AG_INITCURPC:        // pointer to program counter
          pCURPC = static_cast<UL32*>(vp);
          break;
        case AG_INITDOEVENTS:     // DoEvents function pointer
//--- this is no longer relevant due to uVision's threading.
          break;
        case AG_INITUSRMSG:       // Registered Message for SendMessage
          Uv2Msg = PtrToUint(vp); // (Serial-Window, TextOut messages)
          break;
        case AG_INITCALLBACK:     // pointer to callback function
          pCbFunc = static_cast<pCBF>(vp);          // call-back function of s166
          break;
        case AG_INITFLASHLOAD:    // Prepare for Flash Download
          nE = static_cast<U32>(InitProgrammer());
          if (nE) break;
          nE = static_cast<U32>(InitDevFlash());    // Initialize Flash, Erase
          break;
        case AG_STARTFLASHLOAD:   // S8051 says 'Ready for Flash DownLoad'
          nE = static_cast<U32>(FlashLoad());       // Flash-Download, write, verify
          break;
        default:
          break;
      }
      break;

    case AG_GETFEATURE:           // uVision2 wants details about features...
      switch (nCode & 0x00FF)  {
        case AG_F_MEMACCR:    nE = supp.MemAccR; break;
        case AG_F_REGACCR:    nE = supp.RegAccR; break;
        case AG_F_TRACE:      nE = supp.hTrace;  break;
        case AG_F_COVERAGE:   nE = supp.hCover;  break;
        case AG_F_PALYZE:     nE = supp.hPaLyze; break;
        case AG_F_MEMMAP:     nE = supp.hMemMap; break;
        case AG_F_RESETR:     nE = supp.ResetR;  break;

      default:
        break;
      }
      break;

    case AG_EXECITEM:             // execute various commands
      switch (nCode & 0x00FF)  {
        case AG_UNINIT:           // Clean up target system settings
          ExitProgrammer ();      // Stop Programmer
          PlayDead = 1;           // mark target as disconnected.
          CloseAllDlg();          // close all open dialogs
          StopTarget();           // shutdown target & communication
          WriteMonParams (pdbg->TargArgs, _countof(pdbg->TargArgs));  // update argument string
          break;

        case AG_RESET:            // perform a reset on the target system
          if ((PlayDead) || (hMfrm == nullptr)) break;  // target is disconnected
          ResetTarget();          // reset the target
          pCbFunc(AG_CB_EXECCMD, execmd); // dasm $
          NumRecs = 0;            // clear 'number of trace records'
          break;

        default:
          break;
      }
      break;

    default:
      break;
  }

  if (PlayDead)  {       // driver is disconnected
    StopTarget();        // shut down driver
    PostMessage (hMfrm, Uv2Msg, MSG_UV2_TERMINATE, 0);  // unload this driver
  }

  return (nE);
}



/*
 * Memory Interface functions
 */

static DWORD ReadMem (DWORD nAdr, BYTE *pB, DWORD nMany)  {
  DWORD n1, ErrAdr = 0;

  while (nMany != 0)  {
    const DWORD n = (nMany > 256) ? 256 : nMany;
    nMany -= n;
    switch (nAdr >> 24)  {      // extract mSpace from address.
      case amDATA:
        if ((nAdr & 0xFF) >= 0x80)  {  // Sfr-area: D:0x80...D:0xFF
          if (((nAdr & 0xFF) + n) > 0x100) n1 = 0x100 - (nAdr & 0xFF);
          else                             n1 = n;
          ErrAdr = ReadSFR (pB, nAdr & 0xFF0000FFU, n1);
        }
        else  {                     // Data-area: D:0x00...D:0x7F
          if (((nAdr & 0xFF) + n) > 0x100) n1 = 0x100 - (nAdr & 0xFF);
          else                             n1 = n;
          ErrAdr = ReadData (pB, nAdr & 0xFF0000FFU, n1);
        }
        break;

      case amIDATA:                 // Idata
        if (((nAdr & 0xFF) + n) > 0x100) n1 = 0x100 - (nAdr & 0xFF);
        else                             n1 = n;
        ErrAdr = ReadIdata (pB, nAdr & 0xFF0000FFU, n1);
        break;

      case amXDATA:                 // Xdata
      case amPDATA:
        ErrAdr = ReadXdata (pB, nAdr, n);
        break;

//    case amCODE:                  // Code
      default:                      // anything else assumed to be c-address
        ErrAdr = ReadCode (pB, nAdr, n);
        break;
    }
    pB   += n;
    nAdr += n;
  }

  return (ErrAdr);
}

static DWORD WriteMem(DWORD nAdr, const BYTE* pB, DWORD nMany)
{
    DWORD       n1, ErrAdr = 0;
    const BYTE* p1 = pB;                   // save values for caching...
    const DWORD nA = nAdr;
    const DWORD nM = nMany;

    DWORD n = nAdr >> 24;
    if((n == amDATA) || (n == amIDATA)) { // write to data/idata
        if((nAdr & 0xFF) < 0x20)  {       // register bank area
            RegUpToDate = 0;                 // regs might be written...
        }
    }

    while (nMany != 0)  {
        n = (nMany > 256) ? 256 : nMany;
        nMany -= n;
        switch (nAdr >> 24)
        {
            case amDATA:
                if ((nAdr & 0xFF) >= 0x80)  {
                    if (((nAdr & 0xFF) + n) > 0x100) n1 = 0x100 - (nAdr & 0xFF);
                    else                             n1 = n;
                    ErrAdr = WriteSFR(pB, nAdr & 0xFF0000FFU, n1);
                }
                else  {
                    if (((nAdr & 0xFF) + n) > 0x100) n1 = 0x100 - (nAdr & 0xFF);
                    else                             n1 = n;
                    ErrAdr = WriteData(pB, nAdr & 0xFF0000FFU, n1);
                }
                break;

            case amIDATA:
                if (((nAdr & 0xFF) + n) > 0x100) n1 = 0x100 - (nAdr & 0xFF);
                else                             n1 = n;
                ErrAdr = WriteIdata(pB, nAdr & 0xFF0000FFU, n1);
                break;

            case amXDATA:
            case amPDATA:
                ErrAdr = WriteXdata(pB, nAdr, n);
                break;

            case amCODE:
                ErrAdr = WriteCode(pB, nAdr, n);
                break;

            default:
                ErrAdr = nAdr;
                break;
        }
        pB   += n;
        nAdr += n;
    }

    if (ErrAdr == 0)  {                 // write was Ok.
        MyCacheData (p1, nA, nM);           // put write data into cache
    }
    else  {                             // write failed...
        ClearCaR (nA, nM);                // invalidate cache range
    }

    return (ErrAdr);
}


/*
 * Access target memory
 */

EXTERN_C U32 AGDI_DLLEXPORT AG_MemAcc (U16 nCode, UC8 *pB, GADR *pA, UL32 nMany)  {
  U16     nErr;
  UL32      nA;

  if (iRun && !supp.MemAccR)  {     // currently running, can't access.
    return (AG_NOACCESS);           // error: can't access register
  }
  if (PlayDead)  {                  // driver disconnected.
    return (AG_NOACCESS);
  }

  nErr = 0;
  if (hMfrm == nullptr) return (nErr);

  nA = pA->Adr;
  switch (nCode)  {
    case AG_READ:                   // need 'read' permission
      pA->ErrAdr = ReadMem (nA, pB, nMany);
      if (pA->ErrAdr) nErr = AG_RDFAILED;
      break;

    case AG_WRITE:                  // need 'write' permission
      pA->ErrAdr = WriteMem (nA, pB, nMany);
      if (pA->ErrAdr) nErr = AG_WRFAILED;
      break;
                                    // used for Program download
    case AG_WROPC:                  // need 'Read/Execute' permissions
      nA = MAKE_UVISION_CODEPTR(nA);
      pA->ErrAdr = WriteMem (nA, pB, nMany);
      if (pA->ErrAdr) nErr = AG_WRFAILED;
      break;
                                    // used for disassembly etc.
    case AG_RDOPC:                  // need 'Read/Execute' permissions
      nA = MAKE_UVISION_CODEPTR(nA);
      pA->ErrAdr = ReadMem (nA, pB, nMany);
      if (pA->ErrAdr) nErr = AG_RDFAILED;
      break;

    default:
      nErr = AG_INVALOP;            // invalid Operation
      break;
  }

  if(PlayDead == 1) {    // disconnected...
    StopTarget();        // shut down driver
    PostMessage (hMfrm, Uv2Msg, MSG_UV2_TERMINATE, 0);
  }
  return (nErr);
}





/*
 * Read/Write Registers
 */

EXTERN_C U32 AGDI_DLLEXPORT AG_AllReg (U16 nCode, void * /* vp */)  {
  U16     nErr;

  if (iRun && !supp.RegAccR)  {     // currently running, can't access.
    return (AG_NOACCESS);           // error: can't access register
  }
  if (PlayDead) return (AG_NOACCESS);

  nErr = 0;
  switch (nCode)  {                 // which Opcode
    case AG_READ:                   // read register
      GetRegs();
      break;

    case AG_WRITE:
      SetRegs (&REG51);             // write Regs + PC
      RegUpToDate = 1;              // mark regs as 'up to date'
      break;

    default:
      nErr = AG_INVALOP;            // invalid Operation
      break;
  }
  if (PlayDead == 1)  {
    StopTarget();
    PostMessage (hMfrm, Uv2Msg, MSG_UV2_TERMINATE, 0);
  }
  return (nErr);
}


/*
 * Read/Write a single Register
 */

EXTERN_C U32 AGDI_DLLEXPORT AG_RegAcc (U16 nCode, U32 nReg, GVAL *pV)  {
  U16    nErr;

  if (iRun && !supp.RegAccR)  {     // currently running, can't access.
    return (AG_NOACCESS);           // error: can't access register
  }
  if (PlayDead) return (AG_NOACCESS);  // driver is disconnected.

  nErr = 0;
  switch (nCode)  {                 // which Opcode
    case AG_READ:                   // read register
      GetRegs();                    // make sure REG51 is up to date
      switch (nReg)  {
        case nnR0: case nnR1: case nnR2: case nnR3:
        case nnR4: case nnR5: case nnR6: case nnR7:
          pV->uc = REG51.Rn [nReg];
          break;
        case nrA:
          pV->uc = REG51.sfr[REGADDRESS2INDEX(ACC)];
          break;
        case nrDPTR:
          pV->u16 = MAKEWORD(REG51.sfr[REGADDRESS2INDEX(DPL0)], REG51.sfr[REGADDRESS2INDEX(DPH0)]);
          break;
        case mPC:
          pV->u32 = REG51.nPC;      // Note: requires 'amCODE << 24' selector.
          break;
        default:
          break;
      }
      break;

    case AG_WRITE:
      switch (nReg)  {
        case nnR0: case nnR1: case nnR2: case nnR3:
        case nnR4: case nnR5: case nnR6: case nnR7:
          REG51.Rn [nReg & 0x07] = pV->uc;
          SetRegs (&REG51);
          break;
        case nrA:
          REG51.sfr[REGADDRESS2INDEX(ACC)] = pV->uc;
          SetRegs (&REG51);
          break;
        case nrDPTR:
          REG51.sfr[REGADDRESS2INDEX(DPL0)] = LOBYTE(pV->u16);
          REG51.sfr[REGADDRESS2INDEX(DPH0)] = HIBYTE(pV->u16);
          SetRegs (&REG51);
          break;
        case mPC:
          curPC = pV->u32;
          WritePC (curPC);          // to target
          break;
        default:
          break;
      }
      break;

    default:
      nErr = AG_INVALOP;            // invalid Operation
      break;
  }
  if (PlayDead == 1)  {
    StopTarget();
    PostMessage (hMfrm, Uv2Msg, MSG_UV2_TERMINATE, 0);
  }
  return (nErr);
}


/*
 * Memory Attribute access
 */

EXTERN_C U32 AGDI_DLLEXPORT AG_MemAtt (U16 nCode, UL32 nAttr, GADR *pA)  {
  U16     nErr;

  if (iRun && !supp.MemAccR)  {     // currently running, can't access.
    return (AG_NOACCESS);           // error: can't access memory
  }
  if (PlayDead) return (AG_NOACCESS);

  nErr = 0;
  switch (nCode)  {
    case AG_MEMMAP:                 // map memory address range
      pA->ErrAdr = MMapMem (nAttr, pA->Adr, pA->nLen);
      break;

    case AG_GETMEMATT:              // Attribute descriptor is requested
      pA->Adr = PtrToUint(MGetAttr (pA->Adr));  // so return it in pA->Adr.
      break;

    case AG_SETMEMATT:              // not used by uVision.
      break;

    default:
      break;
  }
  if (PlayDead)  {
    StopTarget();
    PostMessage (hMfrm, Uv2Msg, MSG_UV2_TERMINATE, 0);  // terminate driver
  }
  return (nErr);
}

static int SaCBreaks(DWORD set, DWORD nA)  {  // 1:=set, 0:=clear
  AG_Bps* pB;

  abreaks = 0;                    // clear number of address Bp's
  cbreaks = 0;                    // clear number of conditional Bp's
  wbreaks = 0;                    // clear number of access Bp's

  for ( pB = *pBhead ; pB ; pB = pB->next )  {
    if (!pB->enabled) continue;
    switch (pB->type)  {
      case AG_ABREAK: ++abreaks; break;  // increase Nr. of AddressBreaks
      case AG_CBREAK: ++cbreaks; break;  // increase Nr. of CondBreaks
      case AG_WBREAK: ++wbreaks; break;  // increase Nr. of WatchBreaks
      default: break;
    }
  }

  if (set == 0)  {                    // Reset Breaks
    if (SetBrkCount == 0) return (1);
    --SetBrkCount;
  }
  else  {                             // Set Breaks
    if (SetBrkCount != 0) return (1);
    ++SetBrkCount;
  }

//--- when CondBreaks or WatchBreaks are active, skip ExecBreaks
  if (cbreaks || wbreaks) return (1);

  for ( pB = *pBhead ; pB ; pB = pB->next )  {
    if (!pB->enabled) continue;       // Break is disabled
    switch (pB->type)  {
      case AG_ABREAK:                 // Addess-Break
        if (pB->Adr == nA)  {         // g,main address
          pBX = pB;
        }
        if (!SetClrBp (static_cast<int>(set), pB))  {
          return (0);
        }
        break;
      case AG_CBREAK:                 // Conditional-Break
        break;
      case AG_WBREAK:                 // Access-Break
        break;
      default:
        break;
    }
  }
  return (1);
}



/*
 * Is there a Breakpoint at address 'nAdr'
 */

static DWORD IsBreak (DWORD nAdr)  {
  for(AG_Bps* pB = *pBhead ; pB ; pB = pB->next )  {
    if (pB->enabled) {
      if (pB->type == AG_ABREAK)  {
        if (pB->Adr == nAdr)  {
          return (1);               // yes, 'nAdr' has a Bp on it
        }
      }
    }
  }
  return (0);
}


/*
 * Search for Breakpoint at address 'nAdr'
 */

static DWORD BrkWalk (DWORD nAdr)  {
  for(AG_Bps* pB = *pBhead ; pB ; pB = pB->next )  {
    if (!pB->enabled) continue;         // this Break is disabled.
    switch (pB->type)  {
      case AG_ABREAK:                 // Address Break
        if (nAdr == pB->Adr)  {       // match
          if (pB->rcount == 1)  {
            if (pB->cmd)  {           // is command defined
              pCbFunc (AG_CB_EXECCMD, pB->cmd);  // execute command
              return (0);             // continue app when command is defined
            }
            return (1);               // don't continue app
          }
          --pB->rcount;               // decrease counter
          return (0);                 // continue app
        }
        break;

      case AG_CBREAK:                 // Conditional Break, not supported
        break;

      case AG_WBREAK:                 // Watchpoint-Break, not supported
        break;

      default:
        break;
    }
  }
  return (2);                         // no break reason found
}

/*
 * Breakpoint Management
 */

#define BPQX (AG_ATR_BREAK | AG_ATR_BPDIS | AG_ATR_EXECD)

EXTERN_C U32 AGDI_DLLEXPORT AG_BpInfo (U16 nCode, const void* vp)  {
  U32     nA, nAdr;
  U16*    pB = nullptr;
  AG_Bps* pK;

  if (PlayDead) return (0);
  switch (nCode)  {
    case AG_BPDISALL:                  // disable all Exec-Bp's
    case AG_BPKILLALL:                 // kill all Exec-Bp's
      break;
    default:                           // need attribute pointer anyway, set it up here.
      nAdr = static_cast<const GADR*>(vp)->Adr;       // the referred address
      pB   = MGetAttr (nAdr);          // get attribute location for 'nAdr'
      if (!pB) return (0x80000000U);   // failed: unmapped address
      break;
  }

  nA = 0;
  switch (nCode)  {                    // Function code
    case AG_BPQUERY:                   // Query Break
    case AG_BPEXQUERY:                 // Query for 'executed' attribute
      nA = (static_cast<UL32>(*pB) & BPQX);         // Attributes are similar to uVision
      break;

    case AG_BPENABLE:                               // enable breakpoint
      if (*pB & AG_ATR_BPDIS)  {                    // is Bp disabled ?
        *pB = static_cast<U16>((*pB & ~AG_ATR_BPDIS) | AG_ATR_BREAK); // enable it
      }
      break;

    case AG_BPDISABLE:                              // disable breakpoint
      if (*pB & AG_ATR_BREAK)  {                    // is Bp enabled ?
        *pB = static_cast<U16>((*pB & ~AG_ATR_BREAK) | AG_ATR_BPDIS); // disable it
        break;
      }
      break;

    case AG_BPKILL:                                 // kill breakpoint
      *pB = static_cast<U16>(*pB & ~(AG_ATR_BREAK | AG_ATR_BPDIS));
      break;

    case AG_BPSET:                                  // Set breakpoint
      *pB = static_cast<U16>((*pB & ~AG_ATR_BPDIS) | AG_ATR_BREAK);
      break;

    case AG_BPDISALL:                  // disable all Bp's
      pK = *pBhead;                    // head of breakpoint list
      for ( ; pK ; pK = pK->next )  {
        if((pK->type != AG_ABREAK) || (!pK->enabled)) {
          continue;
        }
        ++nA;                          // count number of changed Bp's
        pB = MGetAttr (pK->Adr);             // get attribute location
        if (pB && (*pB & AG_ATR_BREAK))  {   // is Bp enabled ?
          *pB = static_cast<U16>((*pB & ~AG_ATR_BREAK) | AG_ATR_BPDIS); // disable it
        }
      }
      break;

    case AG_BPKILLALL:                 // kill all Bp's
      pK = *pBhead;                    // head of breakpoint list
      for ( ; pK ; pK = pK->next )  {
        if (pK->type != AG_ABREAK) continue;
        ++nA;                          // count number of changed Bp's
        pB = MGetAttr (pK->Adr);       // get attribute location
        if (pB)  {
          *pB = static_cast<U16>(*pB & ~(AG_ATR_BREAK | AG_ATR_BPDIS));
        }
      }
      break;

    default:
      break;
  }

  if (PlayDead == 1)  {
    StopTarget();
    PostMessage (hMfrm, Uv2Msg, MSG_UV2_TERMINATE, 0);
  }
  return (nA);
}


EXTERN_C AGDI_DLLEXPORT AG_Bps* AG_BreakFunc (U16 nCode, U16 /* n1 */, const GADR* pA, AG_Bps* pBp)  {
  U16* pB = nullptr;

  if (PlayDead)
    return(nullptr);                // driver is disconnected

  if (nCode != 5)  {                // not BpAccept function
    pB = MGetAttr (pA->Adr);        // get attribute location
    if (!pB) return (nullptr);      // invalid, cancel
  }

  switch (nCode)  {
    case 2:                         // Notification: 'pB' will be unlinked
      *pB &= ~(AG_ATR_BREAK | AG_ATR_BPDIS);
      break;
    case 3:                          // not used.
    case 6:                          // not used.
    case 7:                          // not used.
    case 8:                          // not used.
      break;
    case 1:                          // Notification: 'pB' will be linked
    case 4:                          // 'pB->enabled' may have changed
      if (pBp->enabled)  {
        *pB = static_cast<U16>((*pB & ~AG_ATR_BPDIS) | AG_ATR_BREAK); // enable it
      }
      else  {
        *pB = static_cast<U16>((*pB & ~AG_ATR_BREAK) | AG_ATR_BPDIS); // disable it
      }
      break;

    case 5:                          // Bp-accept function
      if (pBp->type == AG_WBREAK)  { // read/write access Bp's are not supported here.
        return (nullptr);
      }
      break;

    default:
      break;
  }

  if (PlayDead == 1)  {
    StopTarget();
    PostMessage (hMfrm, Uv2Msg, MSG_UV2_TERMINATE, 0);
  }
  return (pBp);
}



static const char szGWW[] =
    "Application does not run in real time because " \
    "of conditional breakpoint(s)!\n";


/*
 * Execute until 'nAdr' is reached or some other Bp fires.
 */

static void GoUntil (DWORD nAdr)  {     // 0xFFFFFFFF := go forever
  AG_Bps abp;

  pBX = nullptr;                     // clear temporary Bp pointer.
  if (StopRun)  {                 // Stop-Button pressed
    StopRun = 0;
    return;
  }

  if (cbreaks || wbreaks)  {      // cond/watch break(s): need to single step
    txtout (szGWW);               // give the user a note about it.
  }
  else  {                         // Go, setup Bp's first
    if (IsBreak (curPC))  {       // Bp at current PC, step over it.
      Step();
    }
    if (!SaCBreaks (1, nAdr))  {  // set breakpoints (1:=set, 0:=clear)
      StopRun = 0;                // write breakcode(s) failed.
      return;                     // cancel Go
    }
  }

  while (!StopRun)  {
    if (IsBreak (nAdr) || cbreaks || wbreaks)  {
      Step();                    // One Step, then ReadPC()
      Invalidate();              // registers, d/i-caches
      if (BrkWalk (curPC) == 0x01) break;
      if (IsBreak (curPC)) continue;
    }
    if (cbreaks || wbreaks)  {
      continue;                  // continue stepping with c/wbreaks
    }

    abp.enabled = 0;
    if (pBX == nullptr)  {        // setup a temporary Bp
      if (nAdr != 0xFFFFFFFFU)  { // example: g,main
        pBX = &abp;
        abp.type    = AG_ABREAK;
        abp.enabled = 1;
        abp.rcount  = 1;
        abp.ocount  = 1;
        abp.Adr     = nAdr;
        if (!SetClrBp (1, pBX)) { // setup temporary break
          return;
        }
      }
    }

    GoCmd();

    if (abp.enabled)  {           // clear temporary break
      SetClrBp (0, &abp);
      pBX = nullptr;
    }

    if (PlayDead) return;
    curPC = ReadPC();
    if (BrkWalk (curPC)) break;
  }

  SaCBreaks (0, nAdr);            // set breakpoints (1:=set, 0:=clear)
}



/*
 * Go/Step/Stop commands
 */

EXTERN_C AGDI_DLLEXPORT U32 AG_GoStep (U16 nCode, U32 nSteps, const GADR* pA)  {
  U32     nE;

  if (PlayDead) return (0);           // driver is disconnected.

  nE = 0;                             // clear error code
  switch (nCode)  {
    case AG_STOPRUN:                  // Stop Go/Step.
      StopRun = 1;
      nE = StopExec();                // nE:  1=Stopped, 0=executing
      if (nE == 1) iRun = 0;          // stopped.
      if (!iRun)  {
        Invalidate();                 // registers, caches, etc.
      }
      NumRecs = 0;                    // clear 'number of trace records'
      break;

    case AG_NSTEP:                    // execute 'nSteps' instruction steps
      iRun = 1;                       // 'executing'
      for ( ; nSteps != 0 ; --nSteps )  {
        if (StopRun) break;           // Stop-Button was pressed
        if (Step() != 0x01) break;    // 0x01 means 'Ok.'
      }
      StopRun = 0;                    // clear Stop-Button flag.
      Invalidate();                   // registers, caches, etc.
      iRun = 0;                       // clear 'executing' flag
      NumRecs = 0;                    // clear 'number of trace records'
      break;

    case AG_GOTILADR:                 // run til 'pA->Adr' or some Bp,
      iRun = 1;                       // whichever comes first
      GoUntil (pA->Adr);
      StopRun = 0;
      Invalidate();                   // registers, caches, etc.
      iRun = 0;                       // clear 'executing' flag
      NumRecs = 0;                    // clear 'number of trace records'
      break;

    case AG_GOFORBRK:                 // run forever or till some Bp reached.
      iRun = 1;
      GoUntil (0xFFFFFFFFU);          // means 'go forever'
      StopRun = 0;
      Invalidate();                   // registers, caches, etc.
      iRun = 0;                       // clear 'executing' flag
      NumRecs = 0;                    // clear 'number of trace records'
      break;

    default:
      break;
  }

  if (PlayDead)  {                    // target not connected
    StopTarget();                     // shut it down
    PostMessage (hMfrm, Uv2Msg, MSG_UV2_TERMINATE, 0); // terminate ourselves
  }
  else  {
    GetRegs();                        // get target registers
    *pCURPC = REG51.nPC;              // let uVision2 know about PC...
  }
  return (nE);
}

/*
 * Serial Window read/write
 */

EXTERN_C AGDI_DLLEXPORT U32 AG_Serial (U16 nCode, U32 nSerNo, U32 nMany, void *vp)  {
  struct SerAS  va;

  if (PlayDead) return (0);    // driver is disconnected.
  va.n1 = nSerNo;              // 0:=Serial #1,  1:=Serial #2
  va.n2 = nMany;               // number of items to write

  switch (nCode)  {
    case AG_SERBOUT:           // write nMany bytes to Serial #1 Window of uVision2
      va.v.pS = static_cast<SC8*>(vp);  // content
      SendMessage (hMfrm, Uv2Msg, MSG_UV2_SERBOUT, reinterpret_cast<LPARAM>(&va));
      break;

    case AG_SERWOUT:           // write nMany words to Serial #1 Window
//    va.v.pW = static_cast<U16*>(vp);  // content
//    SendMessage (hMfrm, Uv2Msg, MSG_UV2_SERWOUT, (LPARAM) &va);
      break;

    case AG_SERXIN:            // Key was pressed in Serial Window #1 or #2
//    WriteToSerPort (nSerNo, static_cast<char*>(vp), nMany); // Output to serial interface
//    nSerno := Window number (#0 or #1)
//    vp     := &character(s) to send
//    nMany  := number of characters to send
      break;

    default:
      break;
  }
  return (0);
}


#if 0  // Example on how to output a string to uVision's serial window #1:
  char szSerTxt[] = "this should appear in serial window #1";
  AG_Serial (AG_SERBOUT, 0, sizeof (szSerTxt) - 1, (void *) szSerTxt);
#endif



/*
 * Calculate Trace History
 */

#define MAXTRC   256                // assume 1024 (use a power of two value)

static DWORD TracePC [MAXTRC];      // trace code-address buffer


static int CalcNumRecords()  {
  if (NumRecs == 0)  {              // seems to be invalidated

//---just a test vector:
    TracePC [0] = 0x0000000;        // Slot[0] not used.
    TracePC [1] = 0x0000047;
    TracePC [2] = 0x0000049;
    TracePC [3] = 0x000004B;
    TracePC [4] = 0x000004C;
    TracePC [5] = 0x0000050;
    TracePC [6] = 0x0000050;
    TracePC [7] = 0x0000050;
    TracePC [8] = 0x0000050;
    TracePC [9] = 0x0000050;
    TracePC[10] = 0x0000050;
    NumRecs = 10;                   // assume 10 entries

//  Calc:
//  TracePC [0] := unused
//  TracePC [1] := newest code address
//    ..
//  TracePC [n] := oldest code address  ( n >= 0, n < MAXTRC )
  }
  return (NumRecs);
}


/*
 * Trace-History access function:
 *   - the first call is always nCode 3 for 'query number of records'
 *   - then nCode 0 / 1 are activated.
 * Note: Emulator hardware must support tracing
 */

EXTERN_C AGDI_DLLEXPORT U32 AG_HistFunc (U32 nCode, I32 indx, I32 dir, void * /* vp */)  {
  U32     nR = 0;

  if(PlayDead)
  {
    StopTarget();
    PostMessage (hMfrm, Uv2Msg, MSG_UV2_TERMINATE, 0);
    return (0);
  }

  switch (nCode)  {
    case 0:                   // Function #0: get History index
      if (NumRecs == 0) break;
      if (dir)  {             // dir, 1:=towards older entries
        if (indx < NumRecs)  {
          nR = static_cast<U32>(indx + 1);
        }
      }
      else  {                 // dir, 0:=towards newer entries
        if (indx > 0) nR = static_cast<U32>(indx - 1);
        else          nR = 0;
      }
      break;

    case 1:                   // Function #1: GetTracePC
      if((indx >= 0) && (indx < MAXTRC)) {
        nR = MAKE_UVISION_CODEPTR(TracePC [indx]); // get PC of trace record 'indx'
      }
      break;

    case 2:                   // Function #2: GetHistRegs
// this function should be ignored since regs are not recorded anyway.
      break;                  // get recorded regs of record 'indx'

    case 3:                   // Function #3: get Nr. of entries
      if (iRun) return (0);   // block view-trace while running

//   - Find out the number of instructions based on the TRACE-Sfrs up to,
//     but not including the current PC
//***NOTE: make sure that CalcNumRecords() is as fast as possible since
//         it is called many times due to Billy's OnUpdate() handling.
      NumRecs = CalcNumRecords();
      nR      = static_cast<U32>(NumRecs);      // return-value.
      break;

    default:
      break;
  }

  return (nR);
}

#pragma warning(push)
#pragma warning(disable: 4820)
struct bmsk  {
  U16       mask;
  I32       shft;
};
#pragma warning(pop)

static const struct bmsk mx51[] = {
  { 0x0001,  0, },   // P
  { 0x0002,  1, },   // F1
  { 0x0004,  2, },   // OV
  { 0x0018,  3, },   // RS0, RS1
  { 0x0020,  5, },   // F0
  { 0x0040,  6, },   // AC
  { 0x0080,  7, },   // CY
};

static BYTE           nRmode;
static RG51              uRg;         // shadow registers
//static char        szV [128];         // temporary buffer


/*
 * uVision2 wants to display some registers value...
 */

static void RegGet (RITEM *vp, unsigned nR)  {
  GVAL      gv;
  U32       n, z;

  if (PlayDead) return;                    // target is disconnected...

  switch (nR & 0xF0000000U)  {
    case UPR_NORMAL:                       // Setup Normal Regs
      GetRegs();                           // update our local regs
      uRg    = REG51;                      // capture current regs
      nRmode = 0;                          // normal mode
      *pCURPC = uRg.nPC;                   // C:0xnnnn

      if (PlayDead)  {                     // target died...
        StopTarget();
        PostMessage (hMfrm, Uv2Msg, MSG_UV2_TERMINATE, 0);  // Bye
      }
      return;

    case UPR_HIST:                         // Setup History Regs
//--- currently not supported !
      return;

    default:
      break;
  }

  gv.u64     = 0;
  vp->iDraw = 0;

  switch (vp->nItem)  {
    case 0x00:  case 0x01:  case 0x02:  case 0x03:
    case 0x04:  case 0x05:  case 0x06:  case 0x07:
      gv.uc = uRg.Rn [vp->nItem & 0x07];         // current Rn
x8:   if((gv.uc != vp->v.uc) || (vp->szVal[0] == 0))  {
        StringCchPrintfA(vp->szVal, _countof(vp->szVal), "0x%02X", gv.u32);
      }
      break;
    case 0x10: gv.uc = uRg.sfr[REGADDRESS2INDEX(ACC)];  goto x8;  // A
    case 0x11: gv.uc = uRg.sfr[REGADDRESS2INDEX(B)];    goto x8;  // B
    case 0x12: gv.uc = uRg.sfr[REGADDRESS2INDEX(SP)];   goto x8;  // SP
    case 0x13:                            // DPTR
      gv.u16 = static_cast<U16>((static_cast<U16>(uRg.sfr[REGADDRESS2INDEX(DPH0)]) << 8) | uRg.sfr[REGADDRESS2INDEX(DPL0)]);
      if((gv.u16 != vp->v.u16) || (vp->szVal[0] == 0)) {
        StringCchPrintfA(vp->szVal, _countof(vp->szVal), "0x%04X", gv.u32);
      }
      break;

    case 0x14:                            // 'PC  $'
      if((uRg.nPC != vp->v.u32) || (vp->szVal[0] == 0)) {
        StringCchPrintfA(vp->szVal, _countof(vp->szVal), "0x%04lX", uRg.nPC & 0xFFFF);
        vp->iDraw = 1;
      }
      vp->v.u32 = uRg.nPC;
      vp->iHigh = 0;                      // never highlighted
      return;

    case 0x15:                            // States
    case 0x16:                            // Seconds
// not relevant in target mode
      return;

    case 0x100:                           // PSW
    case 0x101:                           // P
    case 0x102:                           // F1
    case 0x103:                           // OV
    case 0x104:                           // RS
    case 0x105:                           // F0
    case 0x106:                           // AC
    case 0x107:                           // Cy
      if (vp->nItem == 0x100)  {          // PSW
        gv.uc = uRg.sfr[REGADDRESS2INDEX(PSW)];// want PSW only
        goto x8;
      }
//--- want some bit from PSW:
      n = vp->nItem - 0x101U;             // use item-number as index
      z = static_cast<U32>(uRg.sfr[REGADDRESS2INDEX(PSW)] & mx51[n].mask);// masked value, not normalized
      gv.u32 = z >> mx51[n].shft;          // normalize
      if((gv.u32 != vp->v.u32) || (vp->szVal[0] == 0)) { // value changed
        StringCchPrintfA(vp->szVal, _countof(vp->szVal), "%u       ", gv.u32);       // spaces to ease mouse hit.
      }
      break;
    default:
      if((vp->nItem >= 0x80) && (vp->nItem <= 0xFF))
      {
        gv.uc = uRg.sfr[REGADDRESS2INDEX(vp->nItem)];
        goto x8;
      }
      else
      {
          const U16 nItem = HIBYTE(vp->nItem);

          if(nItem >= 0x80)
          {
              n = LOBYTE(vp->nItem);

              if(!n)
              {
                  gv.uc = uRg.sfr[REGADDRESS2INDEX(nItem)];
                  goto x8;
              }
              else if(n != 0xFF)
              {
                  z = (uRg.sfr[REGADDRESS2INDEX(nItem)] & n);// masked value, not normalized

                  unsigned shft;

                  for(shft = 0; (!(n & (1U << shft))); ++shft)
                      ;

                  gv.u32 = z >> shft;          // normalize
                  if((gv.u32 != vp->v.u32) || (vp->szVal[0] == 0)) { // value changed
                      StringCchPrintfA(vp->szVal, _countof(vp->szVal), "%u       ", gv.u32);     // spaces to ease mouse hit.
                  }
              }
          }
      }
      break;
  }

  if (gv.u32 == vp->v.u32)  {             // value unchanged
    if (vp->iHigh)  {                     // item is currently highlighted
      vp->iHigh = 0;                      // clear highlight color
      vp->iDraw = 1;                      // need repaint.
    }
  }
  else  {                                 // value changed
    vp->iHigh = 1;                        // highlight
    vp->iDraw = 1;                        // need repaint
  }
  vp->v.u64 = gv.u64;                     // update 'On Screen' value
}


/*
 * uVision2 wants to change some registers value...
 */

static I32 RegSet (const RITEM *vp, const GVAL *pV)  {
  I32       n;
  U16     val;

  if (nRmode == 1) return (0);       // Error: can't change HistRegs !
  if (PlayDead)    return (0);       // cancel: driver is disconnected...
  if (iRun)        return (0);       // cancel: currently running...

  switch (vp->nItem)  {
    case 0x00:  case 0x01:  case 0x02:  case 0x03:   // R0...R7
    case 0x04:  case 0x05:  case 0x06:  case 0x07:
      uRg.Rn [vp->nItem] = pV->uc;
      break;
    case 0x10:                            // A
      uRg.sfr[REGADDRESS2INDEX(ACC)] = pV->uc;
      break;
    case 0x11:                            // B
      uRg.sfr[REGADDRESS2INDEX(B)] = pV->uc;
      break;
    case 0x12:                            // SP
      uRg.sfr[REGADDRESS2INDEX(SP)] = pV->uc;
      break;
    case 0x13:
      uRg.sfr[REGADDRESS2INDEX(DPL0)] = LOBYTE(pV->u16);
      uRg.sfr[REGADDRESS2INDEX(DPH0)] = HIBYTE(pV->u16);
      break;
    case 0x14:
      if(iRun)
        return(0);                        // $ PC: can't change while running...

      uRg.nPC = MAKE_UVISION_CODEPTR(pV->ul & 0xFFFF);
      break;
    case 0x100:                           // PSW
      uRg.sfr[REGADDRESS2INDEX(PSW)] = pV->uc;
      break;
    case 0x101:                           // P
    case 0x102:                           // F1
    case 0x103:                           // OV
    case 0x104:                           // RS
    case 0x105:                           // F0
    case 0x106:                           // AC
    case 0x107:                           // Cy
      val = pV->uc;
      n   = vp->nItem - 0x101;            // use Item-number for index
      uRg.sfr[REGADDRESS2INDEX(PSW)] &= static_cast<BYTE>(~mx51[n].mask); // clear value
      uRg.sfr[REGADDRESS2INDEX(PSW)] |= static_cast<BYTE>((val << mx51[n].shft) & mx51[n].mask);
      break;
    default:
      if((vp->nItem >= 0x80) && (vp->nItem <= 0xFF))
      {
        uRg.sfr[REGADDRESS2INDEX(vp->nItem)] = pV->uc;
      }
      else
      {
        const U16 nItem = HIBYTE(vp->nItem);

        if(nItem >= 0x80)
        {
            n = LOBYTE(vp->nItem);

            if(!n)
            {
                uRg.sfr[REGADDRESS2INDEX(nItem)] = pV->uc;
            }
            else if(n != 0xFF)
            {
                val = pV->uc;
                uRg.sfr[REGADDRESS2INDEX(nItem)] &= static_cast<BYTE>(~static_cast<unsigned>(n)); // clear value

                unsigned shft;

                for(shft = 0; (!(static_cast<unsigned>(n) & (1U << shft))); ++shft)
                    ;

                uRg.sfr[REGADDRESS2INDEX(nItem)] |= static_cast<BYTE>((val << shft) & n);
            }
        }
      }
      break;
  }

  SetRegs(&uRg);

  if (PlayDead == 1)  {           // target died for some reason...
    StopTarget();
    PostMessage (hMfrm, Uv2Msg, MSG_UV2_TERMINATE, 0); // Asta la vista...
  }

  return (1);                                 // Ok.
}

static HRESULT GetFirstFreeDebugAdapterId(ISmartICE04Devices* pDevices, BSTR* pbstr)
{
    ISmartICE04* pDevice;
    VARIANT_BOOL vbl;
    VARIANT varIndex, varDevice;
    HRESULT hr;

    VariantInit(&varIndex);
    VariantInit(&varDevice);
    V_VT(&varIndex) = VT_I4;

    for(V_I4(&varIndex) = 0; SUCCEEDED(hr = pDevices->get_Item(varIndex, &varDevice)); ++V_I4(&varIndex))
    {
        if(SUCCEEDED(VariantChangeTypeEx(&varDevice, &varDevice, CThreadLocale::getLCID(), 0, VT_DISPATCH)))
        {
            if(V_DISPATCH(&varDevice))
            {
                if(SUCCEEDED(V_DISPATCH(&varDevice)->QueryInterface(IID_ISmartICE04, reinterpret_cast<void**>(&pDevice))))
                {
                    if(SUCCEEDED(hr = pDevice->get_IsInDebugMode(&vbl)))
                    {
                        if(!vbl)
                        {
                            if(SUCCEEDED(hr = pDevice->get_SerialNumber(pbstr)))
                                hr = E_INVALIDARG;
                        }
                    }
                    pDevice->Release();
                }
            }
        }
        VariantClear(&varDevice);

        if(FAILED(hr))
            break;
    }

    return((hr == E_INVALIDARG) ? S_OK : hr);
}

static HRESULT IsDeviceUsable(ISmartICE04Devices* pDevices, BSTR bstr)
{
    VARIANT_BOOL vbl;
    VARIANT varIndex, varDevice;
    HRESULT hr;

    VariantInit(&varIndex);
    VariantInit(&varDevice);

    V_VT(&varIndex) = VT_BSTR;
    V_BSTR(&varIndex) = bstr;

    if(SUCCEEDED(hr = pDevices->get_Item(varIndex, &varDevice)))
    {
        if(SUCCEEDED(VariantChangeTypeEx(&varDevice, &varDevice, CThreadLocale::getLCID(), 0, VT_DISPATCH)))
        {
            if(V_DISPATCH(&varDevice))
            {
                ISmartICE04* pDevice = nullptr;

                if(SUCCEEDED(V_DISPATCH(&varDevice)->QueryInterface(IID_ISmartICE04, reinterpret_cast<void**>(&pDevice))))
                {
                    if(SUCCEEDED(hr = pDevice->get_IsInDebugMode(&vbl)))
                    {
                        hr = (vbl) ? S_FALSE : S_OK;
                    }
                    pDevice->Release();
                }
            }
        }
    }
    else if(hr == E_INVALIDARG)
    {
        hr = S_FALSE;
    }

    VariantClear(&varDevice);
    return(hr);
}

static HRESULT GetDebugAdapterId(ISmartICE04Devices* pDevices, BSTR* pDeviceId)
{
    HRESULT hr;

    *pDeviceId = nullptr;

    if(!(MonConf.Opt & (DA_ASKUSER | DA_SPECIFIC)))
    {
        if(FAILED(hr = GetFirstFreeDebugAdapterId(pDevices, pDeviceId)))
            return(hr);

        if(*pDeviceId)
            return(S_OK);
    }

    if(MonConf.Opt & DA_SPECIFIC)
    {
        if(FAILED(hr = BSTRFromStringA(MonConf.idDevice, pDeviceId)))
            return(hr);

        if(S_OK == (hr = IsDeviceUsable(pDevices, *pDeviceId)))
            return(S_OK);

        SysFreeString(*pDeviceId);
        *pDeviceId =  nullptr;

        if(FAILED(hr))
            return(hr);
    }

    if(S_OK != (hr = CDeviceList::DoModal(nullptr)))
        return(hr);

    return(BSTRFromStringA(MonConf.idDevice, pDeviceId));
}
